package com.yizhi.esearch.application.util;

import com.yizhi.core.application.context.ContextHolder;
import com.yizhi.esearch.application.constant.Constant;
import com.yizhi.esearch.application.pojo.biz.*;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DateFormatUtils;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.aggregations.AbstractAggregationBuilder;
import org.elasticsearch.search.aggregations.AggregationBuilders;
import org.elasticsearch.search.aggregations.bucket.histogram.DateHistogramAggregationBuilder;
import org.elasticsearch.search.sort.SortBuilders;
import org.elasticsearch.search.sort.SortOrder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.elasticsearch.core.ElasticsearchRestTemplate;
import org.springframework.data.elasticsearch.core.SearchHit;
import org.springframework.data.elasticsearch.core.SearchHits;
import org.springframework.data.elasticsearch.core.mapping.IndexCoordinates;
import org.springframework.data.elasticsearch.core.query.NativeSearchQueryBuilder;
import org.springframework.data.elasticsearch.core.query.Query;
import org.springframework.util.CollectionUtils;

import javax.validation.constraints.NotNull;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.time.ZoneId;
import java.util.*;

/**
 * @Author: shengchenglong
 * @Date: 2018/12/25 15:19
 */
public class QueryHelper {

    private static final Logger LOGGER = LoggerFactory.getLogger(QueryHelper.class);

    public static final List<String> SEARCH_SETTER_METHOD_NAME = new ArrayList<>();
    public static final List<String> SEARCH_GETTER_METHOD_NAME = new ArrayList<>();
    public static final Map<Class, String> VISIBLE_RANGE_TYPE = new HashMap<>();
    public static final Map<String, Class> VISIBLE_RANGE_CLASS = new HashMap<>();


    static {
        SEARCH_SETTER_METHOD_NAME.add("setName");
        SEARCH_SETTER_METHOD_NAME.add("setKeyword");

        SEARCH_GETTER_METHOD_NAME.add("getName");
        SEARCH_GETTER_METHOD_NAME.add("getKeyword");

        VISIBLE_RANGE_TYPE.put(EsAssignment.class, Constant.BIZ_TYPE_VISIBLE_ASSIGNMENT);
        VISIBLE_RANGE_TYPE.put(EsCaseLibrary.class, Constant.BIZ_TYPE_VISIBLE_CASE_LIBRARY);
        VISIBLE_RANGE_TYPE.put(EsCourse.class, Constant.BIZ_TYPE_VISIBLE_COURSE);
        VISIBLE_RANGE_TYPE.put(EsExam.class, Constant.BIZ_TYPE_VISIBLE_EXAM);
        VISIBLE_RANGE_TYPE.put(EsStudentCase.class, Constant.BIZ_TYPE_VISIBLE_STUDENT_CASE);
        VISIBLE_RANGE_TYPE.put(EsTrainingProject.class, Constant.BIZ_TYPE_VISIBLE_TRAINING_PROJECT);

        VISIBLE_RANGE_CLASS.put(Constant.BIZ_TYPE_VISIBLE_ASSIGNMENT, EsAssignment.class);
        VISIBLE_RANGE_CLASS.put(Constant.BIZ_TYPE_VISIBLE_CASE_LIBRARY, EsCaseLibrary.class);
        VISIBLE_RANGE_CLASS.put(Constant.BIZ_TYPE_VISIBLE_COURSE, EsCourse.class);
        VISIBLE_RANGE_CLASS.put(Constant.BIZ_TYPE_VISIBLE_EXAM, EsExam.class);
        VISIBLE_RANGE_CLASS.put(Constant.BIZ_TYPE_VISIBLE_STUDENT_CASE, EsStudentCase.class);
        VISIBLE_RANGE_CLASS.put(Constant.BIZ_TYPE_VISIBLE_TRAINING_PROJECT, EsTrainingProject.class);
    }

    /**
     * 上下文 租户隔离 queryBuilder
     *
     * @return
     */
    public static QueryBuilder getCommonQueryBuilderForContext() {
        // 站点 id，企业 id queryBuilder
        QueryBuilder queryBuilder = QueryBuilders.boolQuery()
                .must(QueryBuilders.termQuery("deleted", Constant.DELETED_NO))
                .must(QueryBuilders.termQuery("status", Constant.STATUS_UP))
                .must(QueryBuilders.termQuery("companyId", ContextHolder.get().getCompanyId()))
                .must(QueryBuilders.termQuery("siteId", ContextHolder.get().getSiteId()));
        return queryBuilder;
    }

    /**
     * 可见范围 queryBuilder
     *
     * @return
     */
    public static QueryBuilder getCommonQueryBuilderForVisibleRange(ElasticsearchRestTemplate elasticsearchRestTemplate, String indexName, Class clazz) {

        // 全平台可见 visible_range_public
        QueryBuilder queryBuilderForVRP = QueryBuilders.boolQuery()
                .must(QueryBuilders.termQuery("visibleRange", Constant.VISIBLE_RANGE_PUBLIC));

        // visible_range_relationIds
        QueryBuilder queryBuilderForVRR = queryBuilderForVRR(elasticsearchRestTemplate, indexName, clazz);

        // 1. 如果没有可见到的 biz
        if (queryBuilderForVRR == null) {
            return queryBuilderForVRP;
        }

        // 2. 指定范围可见 visible_range_appoint
        QueryBuilder queryBuilderForVRA = QueryBuilders.boolQuery()
                .must(QueryBuilders.termQuery("visibleRange", Constant.VISIBLE_RANGE_APPOINTED))
                .must(queryBuilderForVRR);

        // 组合到一起 visible_range:  appoint or public
        QueryBuilder queryBuilder = QueryBuilders.boolQuery()
                .should(queryBuilderForVRA)
                .should(queryBuilderForVRP);
        LOGGER.info("visible_range queryBuilder----------: " + queryBuilder.toString());
        return queryBuilder;
    }


    /**
     * 获取基础 queryBuilder
     *
     * @return
     */
    public static QueryBuilder getCommonQueryBuilder(ElasticsearchRestTemplate elasticsearchRestTemplate, String indexName, Class clazz) {
        QueryBuilder queryBuilder = QueryBuilders.boolQuery()
                .must(getCommonQueryBuilderForContext())
                .must(getCommonQueryBuilderForVisibleRange(elasticsearchRestTemplate, indexName, clazz));
        return queryBuilder;
    }

    /**
     * OR 条件连接的 queryBuilder
     * <p>
     * TODO 动态字段类型分析
     *
     * @param object
     * @return
     */
    public static QueryBuilder getMustQueryBuilderFromObject(@NotNull Object object) {
        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();

        Method getMethod;
        String name;
        Object value;
        Class clazz = object.getClass();

        try {
            for (String getter : SEARCH_GETTER_METHOD_NAME) {
                getMethod = clazz.getMethod(getter);
                if (getMethod != null) {
                    name = getter.substring(3).toLowerCase();
                    value = getMethod.invoke(object);
                    queryBuilder.should(QueryBuilders.matchQuery(name, value));
                }

            }
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        } catch (Exception e) {
            e.printStackTrace();
            LOGGER.error("", e);
        }
        return queryBuilder;
    }

    /**
     * 设置时区
     *
     * @return
     */
    public static AbstractAggregationBuilder getTimeAggreation() {
        // 取一个名字在改聚合功能器上
        DateHistogramAggregationBuilder builder = AggregationBuilders.dateHistogram("date_transfer")
                .timeZone(ZoneId.of("Asia/Shanghai"))
                .interval(24 * 60 * 60)
                .field("releaseTime")
                .offset("+8h");
//                .minDocCount(0);

        return builder;
    }

    /**
     * 根据关键词搜索分页
     *
     * @param keyword  目前仅支持 name 属性
     * @param pageNo
     * @param pageSize
     * @param clazz
     * @return
     */
    public static Query commonPageSearchQueryForKeyword(String keyword, Class clazz, Integer pageNo, Integer pageSize, String sortFiled, Boolean isAsc, ElasticsearchRestTemplate elasticsearchRestTemplate) {
        try {
            Object object = clazz.newInstance();
            Method method;

            try {
                for (String methodName : SEARCH_SETTER_METHOD_NAME) {
                    method = clazz.getMethod(methodName, String.class);
                    method.invoke(object, keyword);
                }
            } catch (NoSuchMethodException e) {
                LOGGER.info("-----------全局搜索，Class：{} ,没有方法：{}-----------", clazz.toString(), e.getMessage());
            }

            BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery()
                    .must(getMustQueryBuilderFromObject(object));

            QueryBuilder queryBuilderTimeLimit = specificQueryBuilder(clazz);
            if (null != queryBuilderTimeLimit) {
                queryBuilder.must(queryBuilderTimeLimit);
            }

            Query searchQuery = new NativeSearchQueryBuilder()
//                    .addAggregation(getTimeAggreation())
                    .withQuery(queryBuilder)
                    .withFilter(getCommonQueryBuilder(elasticsearchRestTemplate, VISIBLE_RANGE_TYPE.get(clazz), clazz))
                    // page从0开始
                    .withPageable(PageRequest.of(pageNo - 1, pageSize))
                    // notice: 先按评分倒排，再按业务要求排序
                    .withSort(SortBuilders.scoreSort().order(SortOrder.DESC))
                    //增加unmappedType  防止某type找到相应字段
                    .withSort(isAsc ? SortBuilders.fieldSort(sortFiled).order(SortOrder.ASC).unmappedType("long") : SortBuilders.fieldSort(sortFiled).order(SortOrder.DESC).unmappedType("long"))
                    .build();
            return searchQuery;
        } catch (InstantiationException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * 组建时间范围 queryBuilder
     *
     * @param clazz
     * @return
     */
    public static QueryBuilder specificQueryBuilder(Class clazz) {
        QueryBuilder result = null;
        // 调研，时间内
        if (clazz.equals(EsResearch.class)) {
            QueryBuilder queryBuilderStart = QueryBuilders.rangeQuery("startTime")
//                    .format("yyyy-MM-dd")
                    .lte(getToday());

            QueryBuilder queryBuilderEnd = QueryBuilders.rangeQuery("endTime")
//                    .format("yyyy-MM-dd")
                    .gte(getToday());

            result = QueryBuilders.boolQuery()
                    .must(queryBuilderStart)
                    .must(queryBuilderEnd);
        }
        // 课程，排除营销课程
        else if (clazz.equals(EsCourse.class)) {
            result = QueryBuilders.boolQuery()
                    .must(QueryBuilders.termsQuery("source", "1", "2", "3", "5", "6"));
        }
        // 营销课程，排除课程
        else if (clazz.equals(EsMarketingCourse.class)) {
            result = QueryBuilders.boolQuery()
                    .must(QueryBuilders.termQuery("source", "4"));
        }

        return result;
    }

    /**
     * 全局搜索
     *
     * @param keyword
     * @param clazz
     * @param template
     * @return
     */
    public static List globalSearch(String keyword, Class clazz, ElasticsearchRestTemplate template) {
        int pageSize = 0;
        String sortField = null;
        boolean isAsc = true;
        // 课程、项目 最多显示4条、按上架时间倒排
        if (clazz.equals(EsCourse.class) || clazz.equals(EsTrainingProject.class) || clazz.equals(EsMarketingCourse.class)) {
            pageSize = 4;
            sortField = "releaseTime";
            isAsc = false;
        } else if (clazz.equals(EsExam.class) || clazz.equals(EsAssignment.class) || clazz.equals(EsResearch.class)) {
            pageSize = 2;
            sortField = "endTime";
        } else if (clazz.equals(EsStudentCase.class)) {
            pageSize = 2;
            sortField = "status";
        } else {
            pageSize = 2;
            sortField = "status";
        }
        Query searchQuery = QueryHelper.commonPageSearchQueryForKeyword(keyword, clazz, 1, pageSize, sortField, isAsc, template);
        SearchHits searchHits = template.search(searchQuery, clazz);
        // 如果没有命中
        if (!searchHits.hasSearchHits()) {
            return null;
        }
        List list = new ArrayList(Math.toIntExact(searchHits.getTotalHits()));
        searchHits.get().forEach(item -> list.add(item));
        return list;
    }

    /**
     * 检查字段类型
     *
     * @param value
     * @return
     */
    public static Boolean checkValue(Object value) {
        return value != null &&
                (value instanceof Byte
                        || value instanceof Short
                        || value instanceof Integer
                        || value instanceof Long
                        || value instanceof Float
                        || value instanceof Double
                        || value instanceof Character
                        || value instanceof Boolean
                        || value instanceof String);
    }

    private static String buildRelationIdsString() {
        StringBuilder sb = new StringBuilder();
        ContextHolder.get().getRelationIds().stream().forEach(e -> sb.append(e).append(" "));
        return sb.toString();
    }

    /**
     * 组装 relationIds 数组
     *
     * @return
     */
    private static Long[] buildRelationIdsArray() {
        Long[] arr = new Long[ContextHolder.get().getRelationIds().size()];

        return arr;
    }

    /**
     * 组装 relationIds 或 QueryBuilder
     *
     * @return
     */
    private static QueryBuilder queryBuilderForVRR(ElasticsearchRestTemplate elasticsearchRestTemplate, String indexName, Class clazz) {
        List<String> bizIds = getBizIds4VisibleRange(elasticsearchRestTemplate, indexName, clazz);
        if (!CollectionUtils.isEmpty(bizIds)) {
            StringBuilder sb = new StringBuilder();
            bizIds.stream().forEach(e -> sb.append(e).append(" "));
            return QueryBuilders.matchQuery("id", sb.toString());
        }
        return null;
    }

    private static String getToday() {
        return DateFormatUtils.format(new Date(), "yyyy-MM-dd");
    }

    public static List<String> getBizIds4VisibleRange(ElasticsearchRestTemplate elasticsearchRestTemplate, String indexName, Class clazz) {
        if (StringUtils.isEmpty(indexName)) {
            return null;
        }
        Query searchQuery = new NativeSearchQueryBuilder()
                .withQuery(QueryBuilders.matchQuery("relationIds", buildRelationIdsString()))
                .withPageable(PageRequest.of(0, 10000))
                .build();


        SearchHits searchHits = elasticsearchRestTemplate.search(searchQuery, clazz, IndexCoordinates.of(indexName));
        // 如果命中了
        if (searchHits.hasSearchHits()) {
            List<String> bizIds = new ArrayList<>();
            List<SearchHit> searchHitList = searchHits.getSearchHits();
            for (SearchHit hit : searchHitList) {
                bizIds.add(hit.getId());
            }
            return bizIds;
        }

        return null;
    }
}
